from django.test import RequestFactory, Client
from django.http import HttpRequest

from multiverse.utils import get_tenant_model


Tenant = get_tenant_model()


class TenantRequestFactory(RequestFactory):
    tenant: Tenant

    def generic(self, *args, **kwargs):
        if "HTTP_HOST" not in kwargs:
            kwargs["HTTP_HOST"] = self.tenant.subdomain

        request = super().generic(*args, **kwargs)
        request.tenant = self.tenant

        return request


class TenantClientMixin(TenantRequestFactory):
    def login(self, **credentials):
        request = HttpRequest()
        request.META['HTTP_HOST'] = self.tenant.subdomain
        request.tenant = self.tenant

        return super().login(**credentials)


class TenantClient(TenantClientMixin, Client):
    pass



from functools import cache
from django.conf import settings


@cache
def get_tenant_database_alias(or_default: str = 'tenant'):
    return getattr(settings, 'TENANT_DATABASE_ALIAS', or_default)


@cache
def get_tenant_database_name(or_default: str = ':memory:'):
    return getattr(settings, 'TENANT_DATABASE_NAME', or_default)



from rest_framework.test import APITestCase

from multiverse.test.cases import TenantTestCaseMixin
from multiverse.test.drf.client import TenantAPIClient


class TenantAPITestCase(TenantTestCaseMixin, APITestCase):
    client_class = TenantAPIClient



import threading

from typing import Any

from django.conf import settings
from django.db import connections

from multiverse.db.backends.utils import get_tenant_database_alias
from multiverse.signals import tenant_changed
from multiverse.utils import guess_tenant_database_name


__thread_local__ = threading.local()


def get_current_tenant():
    return getattr(__thread_local__, 'current_tenant', None)


def set_current_tenant(tenant: Any):
    setattr(__thread_local__, 'current_tenant', tenant)

    alias = get_tenant_database_alias()
    connections[alias].close()  # close any pending tenant connections...

    # We shouldn't alter settings at runtime...
    # https://docs.djangxoproject.com/en/5.0/topics/settings/#altering-settings-at-runtime
    settings.DATABASES[alias]['NAME'] = tenant.database_name if tenant else ''

    tenant_changed.send(sender=tenant.__class__, instance=tenant)


def forget_current_tenant():
    setattr(__thread_local__, 'current_tenant', None)

    alias = get_tenant_database_alias()
    connections[alias].close()  # close any pending tenant connections...

    # We shouldn't alter settings at runtime...
    # https://docs.djangxoproject.com/en/5.0/topics/settings/#altering-settings-at-runtime
    settings.DATABASES[alias]['NAME'] = guess_tenant_database_name()


def get_request():
    return getattr(__thread_local__, 'request', None)


def set_request(request: Any = None):
    return getattr(__thread_local__, 'request', request)



from django.conf import settings
from django.urls import resolve

from multiverse.awareness import set_current_tenant, set_request
from multiverse.utils import guess_tenant_from_request, get_tenant_model


Tenant = get_tenant_model()


class TenantMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        resolver_match = resolve(request.path)
        
        if resolver_match.app_name in getattr(settings, 'SYSTEM_ROUTES', []):
            return self.get_response(request)

        tenant = guess_tenant_from_request(request)
        setattr(request, 'tenant', tenant)

        set_request(request)
        set_current_tenant(tenant)

        return self.get_response(request)



from mywise.models import BaseModel, models


class BaseTenant(BaseModel):
    subdomain = models.CharField(
        max_length=50,
        null=False,
        unique=True
    )

    database_name = models.CharField(
        max_length=50,
        null=False,
        unique=True
    )

    class Meta:
        abstract = True


class Tenant(BaseTenant):
    class Meta(BaseTenant.Meta):
        swappable = "TENANT_MODEL"



import pathlib
from pathlib import Path


SQLITE_EXTENSIONS = [
    '.sqlite',
    '.sqlite3'
]


# def _sanitize_database_name(name: str | Path) -> Path | str:
#     if ':memory:' in name:
#         return name
#
#     name = Path(name)
#
#     if name.suffix not in SQLITE_EXTENSIONS:
#         name = name.with_suffix(SQLITE_EXTENSIONS[-1])
#
#     if name.parent != settings.BASE_DIR:
#         name = settings.BASE_DIR / name
#
#     return name


def create_database_if_not_exists(name: str | Path) -> tuple[str, bool]:
    name = pathlib.Path(name)  # name = _sanitize_database_name(name)
    created = False

    if not name.exists():
        name.touch()
        created = True

    return str(name), created


def drop_database_if_exists(name: str | Path) -> tuple[str, bool]:
    name = pathlib.Path(name)  # name = _sanitize_database_name(name)
    dropped = False

    if isinstance(name, pathlib.Path) and name.exists():
        name.unlink()
        dropped = True

    return str(name), dropped



# Generated by Django 5.0.7 on 2024-07-18 16:03

import uuid
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Tenant',
            fields=[
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created_at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='updated_at')),
                ('deleted_at', models.DateTimeField(blank=True, null=True, verbose_name='deleted_at')),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('subdomain', models.CharField(max_length=50, unique=True)),
                ('database_name', models.CharField(max_length=50, unique=True)),
            ],
            options={
                'abstract': False,
                'swappable': 'TENANT_MODEL',
            },
        ),
    ]



def __require_psycopg():
    try:
        import psycopg
    except ImportError as e:
        raise ImportError(
            'psycopg not installed. '
            'run `pip install django-multiverse[postgres]` or `pip install psycopg`'
        ) from e


__require_psycopg()



import uuid

from django.db import models

from importlib import import_module
from types import ModuleType

from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.management import call_command
from django.db.models import Q
from django.shortcuts import get_object_or_404


from multiverse.db.backends.utils import get_tenant_database_alias, get_tenant_database_name


def is_valid_uuid(val: uuid.UUID | str) -> bool:
    if isinstance(val, uuid.UUID):
        return True

    if not isinstance(val, str):
        return False

    try:
        uuid_obj = uuid.UUID(val)
        return str(uuid_obj) == val
    except ValueError:
        return False


def get_tenant_model():
    """
    Return the User model that is active in this project.
    """
    try:
        return django_apps.get_model(settings.TENANT_MODEL, require_ready=False)
    except ValueError:
        raise ImproperlyConfigured(
            "TENANT_MODEL must be of the form 'app_label.model_name'"
        )
    except LookupError:
        raise ImproperlyConfigured(
            "TENANT_MODEL refers to model '%s' that has not been installed"
            % settings.TENANT_MODEL
        )


def guess_tenant_database_name():
    tenant_database_alias = get_tenant_database_alias()
    tenant_database_settings = settings.DATABASES.get(tenant_database_alias, {})

    # if not found, the fallback will be the last set at settings.DATABASES[<tenant>]['NAME']
    # or :memory: in case even the name doesn't exist. (assuming sqlite ENGINE in this case)
    or_default = tenant_database_settings.get('NAME') or ':memory:'

    return get_tenant_database_name(or_default=or_default)


def guess_tenant_from_request(request):
    """
    Determines the tenant based on the incoming request.

    This function attempts to identify the tenant by examining the subdomain or hostname
    in the request. It follows these steps:

    1. Checks if the tenant subdomain (Header: X-Tenant) is present in the request.
    2. If not, it checks the hostname (Header: Host).
    3. If the application is running in DEBUG mode and the hostname is '127.0.0.1' or 'localhost',
       it returns the default tenant for the development environment.
    4. If none of the above conditions are met, it extracts the subdomain from the hostname.

    Args:
        request: The HTTP request object.

    Returns:
        Tenant: The tenant object corresponding to the subdomain or hostname.

    Raises:
        Http404: If no tenant is found for the given subdomain or hostname.

    """
    Tenant = get_tenant_model()

    if is_tenant_subdomain_in_request(request):
        subdomain = get_tenant_subdomain_from_request(request)
        return get_object_or_404(Tenant, subdomain=subdomain)

    hostname = get_hostname_from_request(request)

    if settings.DEBUG and hostname in ['127.0.0.1', 'localhost']:
        # Failed to guess tenant from request by neither headers: X-Tenant and Host.
        # If the app is running in debug mode, and it's a local IP or hostname,
        # allow getting the default tenant for the development environment.
        default_database_name = guess_tenant_database_name()
        return get_object_or_404(Tenant, database_name=default_database_name)

    subdomain = hostname.split('.')[0]
    return get_object_or_404(Tenant, subdomain=subdomain)


def is_tenant_subdomain_in_request(request, header_name: str = 'X-Tenant'):
    return header_name in request.headers


def get_tenant_subdomain_from_request(request, header_name: str = 'X-Tenant'):
    # split on `:` to remove the port
    if is_tenant_subdomain_in_request(request, header_name):
        return request.headers[header_name]

    return None


def get_hostname_from_request(request):
    # split on `:` to remove the port
    return request.get_host().split(':')[0].lower()


def matches_int_field(field: models.Field, value):
    return isinstance(field, models.AutoField) and isinstance(value, int)


def matches_uuid_field(field: models.Field, value):
    return isinstance(field, models.UUIDField) and isinstance(value, (str, uuid.UUID)) and is_valid_uuid(value)


def get_tenant(lookup: str | uuid.UUID | int):
    Tenant = get_tenant_model()

    pk_field = Tenant._meta.pk

    if matches_int_field(pk_field, lookup) or matches_uuid_field(pk_field, lookup):
        return Tenant.objects.get(pk=lookup)

    where = Q(
        _connector=Q.OR,
        subdomain=lookup,
        database_name=lookup
    )

    return Tenant.objects.get(where)


def get_common_apps(or_default: list[str] | None = None):
    return getattr(settings, 'COMMON_APPS',  or_default or [])


def get_system_apps(or_default: list[str] | None = None):
    return getattr(settings, 'SYSTEM_APPS',  or_default or [])


def get_tenant_apps(or_default: list[str] | None = None):
    return getattr(settings, 'TENANT_APPS',  or_default or [])


def get_db_utils_module(db_alias: str) -> ModuleType | None:
    if db_alias not in settings.DATABASES:
        return None

    db_engine = settings.DATABASES[db_alias]['ENGINE']

    if 'sqlite3' in db_engine:
        return import_module('multiverse.db.backends.sqlite3.utils')

    if 'postgresql' in db_engine:
        return import_module('multiverse.db.backends.postgresql.utils')

    return None


def create_tenant_database(tenant):
    db_alias = get_tenant_database_alias()
    db_utils_module = get_db_utils_module(db_alias)

    if db_utils_module and hasattr(db_utils_module, 'create_database_if_not_exists'):
        db_utils_module.create_database_if_not_exists(tenant.database_name)


def drop_tenant_database(tenant):
    db_alias = get_tenant_database_alias()
    db_utils_module = get_db_utils_module(db_alias)

    if db_utils_module and hasattr(db_utils_module, 'drop_database_if_exists'):
        db_utils_module.drop_database_if_exists(tenant.database_name)


def migrate_tenant_database():
    # Run migrations for the new tenant's database
    call_command('migrate', database=get_tenant_database_alias())



from django.core.management.base import BaseCommand

from multiverse.utils import get_tenant_model, create_tenant_database, migrate_tenant_database

Tenant = get_tenant_model()


class Command(BaseCommand):
    help = 'Creates a tenant for MyWise'

    def add_arguments(self, parser):
        parser.add_argument('subdomain', type=str)
        parser.add_argument('--database-name', type=str, nargs='?')

        parser.add_argument('--create-database', type=bool, default=True)
        parser.add_argument('--migrate', type=bool, default=True)

    def handle(self, *args, **options):
        subdomain = options.get('subdomain')
        database_name = options.get('database_name') or subdomain

        tenant, created = Tenant.objects.get_or_create(
            subdomain=subdomain,
            database_name=database_name
        )

        if options.get('create_database'):
            create_tenant_database(tenant)

        if options.get('migrate'):
            migrate_tenant_database()

        self.stdout.write(
            self.style.SUCCESS(f"Tenant with Domain '{subdomain}' and Database '{database_name}' created successfully")
        )



import psycopg
from psycopg import sql

from django.conf import settings


def _get_connection() -> psycopg.Connection:
    return psycopg.connect(
        dbname="postgres",
        user=settings.DATABASES['default']['USER'],
        password=settings.DATABASES['default']['PASSWORD'],
        host=settings.DATABASES['default']['HOST'],
        port=settings.DATABASES['default']['PORT'],
        autocommit=True,
    )


def _database_exists(cursor, name) -> bool:
    cursor.execute(
        sql.SQL(
            'SELECT datname FROM pg_catalog.pg_database WHERE datname = %(datname)s'
        ),
        params={
            'datname': name
        }
    )

    return cursor.rowcount > 0


def _database_create(cursor, name):
    cursor.execute(
        sql.SQL('CREATE DATABASE {}').format(
            sql.Identifier(name)
        )
    )


def _database_drop(cursor, name):
    cursor.execute(
        sql.SQL('CREATE DATABASE IF EXISTS {}').format(
            sql.Identifier(name)
        )
    )


def create_database_if_not_exists(name: str) -> tuple[str, bool]:
    connection = _get_connection()
    created = False

    with connection.cursor() as cursor:
        if not _database_exists(cursor, name):
            _database_create(cursor, name)
            connection.commit()
            created = True

    connection.close()

    return name, created


def drop_database_if_exists(name: str) -> tuple[str, bool]:
    connection = _get_connection()
    dropped = False

    with connection.cursor() as cursor:
        if _database_exists(cursor, name):
            _database_drop(cursor, name)
            connection.commit()
            dropped = True

    connection.close()

    return name, dropped



from django.db import DEFAULT_DB_ALIAS

from multiverse.test import is_test_environment
from multiverse.utils import get_tenant_database_alias, get_system_apps, get_common_apps
from multiverse.awareness import get_current_tenant


class TenantRouter:
    def app_label_in_apps(self, app_label, apps):
        for app in apps:
            # eg, typical apps: mycustomapp
            if app_label == app:
                return True

            # eg, apps with full name: django.contrib.contenttypes
            if app_label == app.split('.')[-1]:
                return True

        return False

    def db_for(self, mode, model, **hints):
        if is_test_environment():
            return DEFAULT_DB_ALIAS

        app_label = model._meta.app_label

        if self.app_label_in_apps(app_label, get_common_apps()):
            return None

        if self.app_label_in_apps(app_label, get_system_apps()):
            return DEFAULT_DB_ALIAS

        return get_tenant_database_alias()

    def db_for_read(self, model, **hints):
        return self.db_for('read', model, **hints)

    def db_for_write(self, model, **hints):
        return self.db_for('write', model, **hints)

    def allow_relation(self, obj1, obj2, **hints):
        db_obj1 = hints.get('database', None) or self.db_for_read(obj1)
        db_obj2 = hints.get('database', None) or self.db_for_read(obj2)

        return db_obj1 and db_obj2 and db_obj1 == db_obj2

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        if is_test_environment():
            return True

        if self.app_label_in_apps(app_label, get_common_apps()):
            return True

        if self.app_label_in_apps(app_label, get_system_apps()):
            return db == DEFAULT_DB_ALIAS

        return db == get_tenant_database_alias()



from django.test import TestCase
from django.conf import settings

from multiverse.utils import get_tenant_model, guess_tenant_database_name
from multiverse.test.client import TenantClient
from multiverse.test.utls import set_test_environment


Tenant = get_tenant_model()


class TenantTestCaseMixin:
    tenant: Tenant = None
    databases = '__all__'

    def __init__(self, *args, **kwargs):
        # needed here as well due to the action of database migration in a testing environment
        set_test_environment(True)
        super().__init__(*args, **kwargs)

    @classmethod
    def add_allowed_host(cls, host):
        if host not in settings.ALLOWED_HOSTS:
            settings.ALLOWED_HOSTS.append(host)

    @classmethod
    def remove_allowed_host(cls, host):
        if host in settings.ALLOWED_HOSTS:
            settings.ALLOWED_HOSTS.remove(host)

    @classmethod
    def setUpClass(cls):
        set_test_environment(True)

        subdomain = getattr(cls.tenant, 'subdomain', 'test')
        database_name = getattr(cls.tenant, 'database_name', guess_tenant_database_name())

        cls.tenant, _ = Tenant.objects.get_or_create(subdomain=subdomain, database_name=database_name)
        cls.add_allowed_host(cls.tenant.subdomain)

        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

        if cls.tenant:
            cls.remove_allowed_host(cls.tenant.subdomain)
            cls.tenant = None

        set_test_environment(False)

    def _pre_setup(self):
        super()._pre_setup()

        if hasattr(self, 'client') and isinstance(self.client, TenantClient):
            self.client.tenant = self.tenant

    def _post_teardown(self):
        if hasattr(self, 'client') and isinstance(self.client, TenantClient):
            self.client.tenant = None

        super()._post_teardown()


class TenantTestCase(TenantTestCaseMixin, TestCase):
    client_class = TenantClient



def __require_djangorestframework() -> list[str]:
    try:
        import rest_framework
    except ImportError as e:
        raise ImportError(
                'djangorestframework not installed. '
                'run `pip install django-multiverse[drf]` or `pip install djangorestframework`'
                'and add "rest_framework" to your INSTALLED_APPS') \
            from e


__require_djangorestframework()


from multiverse.test.drf.cases import TenantAPITestCase


__all__ = [
    'TenantAPITestCase'
]



def __require_django_q2():
    try:
        import django_q
    except ImportError as e:
        raise ImportError(
            'django-q2 not installed. '
            'run `pip install django-multiverse[django-q2]` or `pip install django-q2`'
        ) from e


__require_django_q2()



from django.core.management.base import BaseCommand

from multiverse.utils import get_tenant, drop_tenant_database


class Command(BaseCommand):
    help = 'Destroys a tenant of MyWise'

    def add_arguments(self, parser):
        parser.add_argument('lookup', type=str)
        parser.add_argument('--drop-database', type=bool, default=False)

    def handle(self, *args, **options):
        lookup = options.get('lookup')
        drop_database = options.get('drop_database')

        tenant = get_tenant(lookup)
        tenant.delete(using='default', hard=drop_database)

        if drop_database:
            drop_tenant_database(tenant)



from django.conf import settings


SETTINGS_TESTING_KEY = 'TESTING'


def is_test_environment():
    return getattr(settings, SETTINGS_TESTING_KEY, False)


def set_test_environment(testing: bool):
    return setattr(settings, SETTINGS_TESTING_KEY, testing)



from importlib import import_module

from django_q.tasks import schedule as django_q_schedule, async_task as django_q_async_task
from django_q.utils import get_func_repr

from multiverse.awareness import get_current_tenant, set_current_tenant
from multiverse.utils import get_tenant


def tenant_aware_func(tenant_id, func, *args, **kwargs):
    tenant = get_tenant(tenant_id)
    set_current_tenant(tenant)

    if isinstance(func, str):
        module_name, func_name = func.rsplit('.', 1)

        module = import_module(module_name)
        func = getattr(module, func_name)

    return func(*args, **kwargs)


def async_task(func, *args, **kwargs):
    tenant = get_current_tenant()

    if not tenant:
        return django_q_async_task(get_func_repr(func), *args, **kwargs)

    return django_q_async_task(
        get_func_repr(tenant_aware_func),
        str(tenant.pk),
        get_func_repr(func),
        *args,
        **kwargs
    )


def schedule(func, *args, **kwargs):
    tenant = get_current_tenant()

    if not tenant:
        return django_q_async_task(get_func_repr(func), *args, **kwargs)

    return django_q_schedule(
        get_func_repr(tenant_aware_func),
        str(tenant.pk),
        get_func_repr(func),
        *args,
        **kwargs
    )
